#' genCov function 
#'
#' This function calculates an additive genetic covariance matrix from dominant genetic marker coded {0, 1}, such as DArT markers. 
#'
#' @param M matrix. A g x m matrix of m markers for g genotypes. Cannot handle missing data as is. Impute with marker score mean for a quick fix! 
#' @param lambda numeric. Matrix bending coefficient to be added to the diagonal of the covariance matrix to recover full rank after centering. Centering projects X onto n-1 dimensional subspace (i.e. X is rank deficient)
#' @param returnH logical. if TRUE the additive by additive epistatic covariance is also returned along with the additive genetic covariance in a list of length 2.  
#' @return Additive and epistatic genetic covariances
#' @details
#' 
#' Calculates the additive genetic covariance, K, based on dominant molecular markers based on method I of Van Raden (2008). 
#' If 'returnH' is set to TRUE, the additive by additive epistatic covariance, H, is also returned in a list with elements 'K' and 'H'. This is calculated according to Martini et al. (2016), but generally follows the methodology of Jiang et al. (2015)
#'
#' @examples
#' library(wheat)
#' data(wheat)
#' 
#' K <- genCov(wheat.X)
#' KH <- genCov(wheat.X, returnH = TRUE)
#' lapply(KH, dim)
#'
#' @keywords Genomic Selection, K-fold cross-validation , Bayer Crop Science (and Monsanto :P)
#' @export
genCov <- function(M, lambda = 0.01, returnH = FALSE){
	if (!all(M %in% c(0, 1))) stop("Expecting dominant coded marker scores, in {0, 1} ")
	p <- colMeans(M)
	Z <- scale(M * 2, scale = FALSE)
	ZZt <- tcrossprod(Z)
	alVar <- (2 * c(crossprod(p, 1-p)))
	K <- ZZt / alVar
	if (returnH) H <- K^2 - tcrossprod(Z^2) / alVar^2 
	if(lambda != 0) {
		diag(K) <- diag(K) + lambda
		if (returnH) diag(H) <- diag(H) + lambda
	}
	if(returnH) return(list(K = K, H = H)) else return(K)
}