#' lreFit function 
#'
#' This function fits three nested models for low resolution epistasis mapping using the EMMREML package to solve the mixed model equations. 
#'
#' @param y numeric. Phenotype response vector of length n
#' @param M matrix. A g x m matrix of m markers for g genotypes. Cannot handle missing data as is. Impute with marker score mean for a quick fix! 
#' @param X matrix. n x p design matrix. Cn include dummy variables (e.g. trials) as well as numeric covariates
#' @param Z matrix. n x g incidence matrix of genotypes
#' @param index1 logical. Boolean vector of length m indicating which columns of M to include in region 1 
#' @param index2 logical. Boolean vector of length m indicating which columns of M to include in region 2 
#' @param verbose logical. Boolean vector of length 1 indicating whether or not information from the model fit should be printed to stdout. 
#' @param returnFullFit logical. Boolean vector of length 1 indicating whether or not the output from all EMMREML model fits should be returned. 
#' @return P-values for a model including the two regions as an additive effect and an interaction. Random effects (BLUPs) for the two regions specified as well as their interaction are also returned   
#' @details
#' 
#' This function fits the low resolution epistasis model proposed by Santantonio Jannink and Sorrells (2018). A low resolution epistasis mapping approach to identify chromosome arm interactions in allohexaploid wheat. Submitted to G3
#' 
#' Briefly, marker scores belonging to two regions indexed by arguments index1 and index2 are isolated from the remaining markers scores. These are used to cumpute additive genetic covarianes for each region and the remaining markers to fit additive kernels for each region while correcting for background additive and epistatic effecs. 
#'
#' Three models are fit, the null model without the regions of interest, the additive model including the regions as additive kernels, and the interaction model that includes the interaction at the two regions. 
#'
#' A nested likelihood ratio test is used to produce a p-value to determine if at least one additive variance componets is greater than zero, as well as if the interaction variance component is greater than zero. 
#'
#' A list with elements 'pvalues' and 'BLUPs' containing pvalues and BLUPs for additive and interraction effects, respectively, is returned. 
#'
#' @examples
#' data(CNLM)
#' 
#'
#' @keywords Low resolution epistasis mapping.
#' @export
lreFit <- function(y, M, X, Z, index1, index2, verbose = FALSE, returnFullFit = FALSE){

	if(!all(colnames(Z) == rownames(M))) stop("Column names of Z must match rownames of M!")
	if (!"EMMREML" %in% rownames(installed.packages())) install.packages("EMMREML")
	require(EMMREML)
	
	if (!all(c(class(index1), class(index2)) %in% "logical")) {
		stop("Please provde a logical vector of length ncol(M)to both 'index1' and 'index2' arguments.")
	}
	if (length(index1) != ncol(M) | length(index2) != ncol(M)) {
		stop ("'index1' and 'index2' must have length equal to the number of columns of M!")
	}
	if (sum(index1 | index2) == ncol(M)) {
		stop ("'index1' and 'index2' contain all markers! Some residual markers must be left over to correcr for background effects.")
	}
	pval <- list()
	blup <- list()

	GminusIndex <- !(index2 | index1)
	if (verbose){
		cat("Number of markers from region 1:", sum(index1), "\n")
		cat("Number of markers from region 2:", sum(index2), "\n")
		cat("Number of markers correcting for background effects:", sum(GminusIndex), "\n")
	}

	M1 <- M[, index1, drop = FALSE]
	M2 <- M[, index2, drop = FALSE]
	
	K1 <- genCov(M1)
	K2 <- genCov(M2)
	K12 <- K1 * K2		

	KHminus <- genCov(M[, GminusIndex], returnH = TRUE)

	zlist <- list(Z1 = Z, Z2 = Z, Z12 = Z, Zminus = Z, Z2minus = Z)
	klist <- list(K1 = K1, K2 = K2, K12 = K12, Kminus = KHminus$K, K2minus = KHminus$H)

	if (verbose) cat("Fitting null model...\n")
	armFitH0 <- emmremlMultiKernel(y = y, X = X, Zlist = zlist[c("Zminus", "Z2minus")], Klist = klist[c("Kminus", "K2minus")])
	
	if (verbose) cat("Fitting additive model...\n")	
	armFitH1add <- emmremlMultiKernel(y = y, X = X, Zlist = zlist[c("Z1", "Z2", "Zminus", "Z2minus")], Klist = klist[c("K1", "K2", "Kminus", "K2minus")])
	pval[["add"]] <- LRT(L1 = armFitH1add$loglik, L0 = armFitH0$loglik, dF = 2)
	cat("    Additive model p-value:", signif(pval[["add"]]), "\n")	

	if (verbose) cat("Fitting interaction model...\n")	
	armFitH1int <- emmremlMultiKernel(y = y, X = X, Zlist = zlist[c("Z1", "Z2", "Z12", "Zminus", "Z2minus")], Klist = klist[c("K1", "K2", "K12", "Kminus", "K2minus")])
	pval[["epi"]] <- LRT(L1 = armFitH1int$loglik, L0 = armFitH1add$loglik, dF = 1)
	cat("    Interaction model p-value:", signif(pval[["epi"]]), "\n")	

	blup[["add1"]] <- armFitH1int$u[grep("K_1", names(armFitH1int$u))]
	blup[["add2"]] <- armFitH1int$u[grep("K_2", names(armFitH1int$u))]
	blup[["epi"]] <- armFitH1int$u[grep("K_3", names(armFitH1int$u))]

	returnL <-list(pvalues = pval, blups = blup)
	if(returnFullFit) returnL <- c(returnL,  list(nullModel = armFitH0, addModel = armFitH1add, intModel = armFitH1int))
	return(returnL)
}