##############################################################################
# This script will run the low resolution chromosome arm epistasis analysis. #
##############################################################################

##############################################################################################
# I highly recommend using Microsoft R open, which takes advantage of the MKL math libraries #
# that are far more efficient and allow for multicore computing that will greatly reduce     #
# the time to fit these models.             https://mran.microsoft.com/open                  #
##############################################################################################

# Uncomment to install the lre package, this only needs to be run ONCE!
# system("R CMD INSTALL lre") 

# load the lre library
library(lre)

# load the data from Santantonio, Jannink, Sorrells. 2018. 
# Prediction of subgenome additive and interaction effects in allohexaploid wheat. Submitted to G3
data(CNLM)

###################
#    Data Prep    #
###################

# make chromosome arm variable, 'chrArm' 
snpInfo$chrArm <- paste0(snpInfo$chrom, snpInfo$arm)

# subset records with plant height (PH) data
pheno <- droplevels(pheno[!is.na(pheno$PH), c("GID", "Trial", "PH")])

# make experimental design and genotype incidence matrices
X <- model.matrix(~ Trial, data = pheno) 
Z <- model.matrix(~ GID - 1, data = pheno)
colnames(Z) <- gsub("GID", "", colnames(Z))

# order genotype row of M based on columns of Z, so that the resulting columns and rows of K 
# correspond to the correct columns of Z
M <- M[colnames(Z), ] 

# make list of chromosome arm pairs
armList <- list(c("4Bs", "4Ds"), c("4Bl", "4Dl"))

###################
#    Model Fit    #
###################

# loop through armList pairs, fit null, additive and interaction model to test for interaction effects
pvals <- list()
blups <- list()
for(i in 1:length(armList)){
	# Save name of regions (i.e. arms)
	iname <- paste(armList[[i]], collapse = "_")

	# create boolean vectors indicating which markers belong to each of the two regions
	arm1 <- colnames(M) %in% snpInfo$snp[snpInfo$chrArm %in% armList[[i]][1]]
	arm2 <- colnames(M) %in% snpInfo$snp[snpInfo$chrArm %in% armList[[i]][2]]

	cat("Fit interaction model for ", iname, "\n")
	# fit the low resolution epistasis model
	fit <- lreFit(y = pheno$PH, M = M, X = X, Z = Z, index1 = arm1, index2 = arm2, verbose = TRUE, returnFullFit = FALSE)

	#extract pvalues and BLUPs from output
	pvals[[iname]] <- fit$pval
	blups[[iname]] <- fit$blups
}

# print p-values
print(pvals)


####################
# Plotting effects #
####################

# loop through blups and plot interaction effect by the product of the additive effects
par(mfrow = c(1, 2)) # plot both on the same window.
for(i in names(blups)){
	# make product of additive effects, get interaction effect and correlation coefficient 
	addProd <- blups[[i]][["add1"]] * blups[[i]][["add2"]]
	int <- blups[[i]][["epi"]]
	rho <- cor(addProd, int)

	inames <- strsplit(i, "_")[[1]]
	# plot the additive product by interaction effect 
	plot(addProd, int, pch = 21, bg = "#0000004D", 
		main = paste(paste(inames, collapse = " x "), "versus", paste(inames, collapse = "*")), 
		 xlab = paste(paste(inames, collapse = " x "), "Additive Product"), 
		 ylab = paste(paste(inames, collapse = "*"), "Interaction"))
	legend("topright", legend = bquote(rho == .(round(rho, 2))), bty = "n")
}

###################
# other intervals #
###################

# we could redefine some other interval. Say splitting the long arms of homeolog 4 into smaller pieces

# break long arm of homeolog 4 into equal sized pieces by number of markers
len4Bl <- sum(snpInfo$chrArm == "4Bl")
half4Bl <- floor(len4Bl / 2)

len4Dl <- sum(snpInfo$chrArm == "4Dl")
half4Dl <- floor(len4Dl / 2)

# get marker names for each part
m4Bl1 <- snpInfo[snpInfo$chrArm == "4Bl", "snp"][1:half4Bl]
m4Bl2 <- snpInfo[snpInfo$chrArm == "4Bl", "snp"][(half4Bl+ 1):len4Bl]

m4Dl1 <- snpInfo[snpInfo$chrArm == "4Dl", "snp"][1:half4Dl]
m4Dl2 <- snpInfo[snpInfo$chrArm == "4Dl", "snp"][(half4Dl+ 1):len4Dl]

# make list of markers in each part
mList <- list(chr4l1 = list(m4Bl1 = m4Bl1, m4Dl1 = m4Dl1), chr4l2 = list(m4Bl2 = m4Bl2, m4Dl2 = m4Dl2))

# loop through 4l parts 1 and 2, fit null, additive and interaction model to test for interaction effects
pvals4l <- list()
blups4l <- list()
for(i in names(mList)){

	# create boolean vectors indicating which markers belong to each of the two regions
	part1 <- colnames(M) %in% mList[[i]][[1]]
	part2 <- colnames(M) %in% mList[[i]][[2]]

	cat("Fit interaction model for ", i, "\n")
	# fit the low resolution epistasis model
	fit <- lreFit(y = pheno$PH, M = M, X = X, Z = Z, index1 = part1, index2 = part2, verbose = TRUE, returnFullFit = FALSE)

	#extract pvalues and BLUPs from output
	pvals4l[[i]] <- fit$pval
	blups4l[[i]] <- fit$blups
}

# print p-values
print(pvals4l)


##################################################
#                                                #
# Below shows how to fit the models in parallel. #
#                                                #
##################################################

########################
# lreFit() in parallel #
########################

# Note: This is for illustration only. If using Microsoft R open this will NOT be faster for only two pairs 

# The model fit can also be run in parallel using the 'doMC' package 
# define a wrapper function for lreFit 
lreArms <- function(arms, snpInfo, y, M, X, Z){
	rL <- list()
	iname <- paste(arms, collapse = "_")
	arm1 <- colnames(M) %in% snpInfo$snp[snpInfo$chrArm %in% arms[1]]
	arm2 <- colnames(M) %in% snpInfo$snp[snpInfo$chrArm %in% arms[2]]
	rL[[iname]] <- lreFit(y = y, M = M, X = X, Z = Z, index1 = arm1, index2 = arm2, verbose = TRUE, returnFullFit = TRUE)
	rL
}

# load 'doMC' library 
library("doMC")

# If using Microsoft R open, set number of MKL threads for each parallel thread. 
# This should not exceed the number of cores / number of parallel threads!
# For example, if you had 8 cores, the maximum value would be setMKLthreads(4) for registerDoMC(cores = 2)
# setMKLthreads(18)

# set number of parallel threads
registerDoMC(cores = 2)

# run each pair in parallel. Uncomment to execute. Not recommended for only the two arms shown here. 
fitList <- foreach(i = 1:length(armList), .combine = c) %dopar% lreArms(arms = armList[[i]], snpInfo = snpInfo, y = pheno$PH, M = M, X = X, Z = Z)

# print pvalues
print(lapply(fitList, "[[", "pval"))


###########################################################################################################
# Uncomment the following block to run all 861 pairs for all four traits.                                 #
# THIS IS NOT RECOMMENDED! FITTING ALL 861 PAIRS COULD TAKE WEEKS, DEPENDING MACHINE, MATH LIBRARIES ETC. #
# The use of ASReml is necessary to fit all pairs in a reasonable timeframe.                              #
###########################################################################################################

# arm <- unique(snpInfo$chrArm)
# arm <- arm[!grepl("inCentro")]
# armList <- combn(arm, 2, simplify = FALSE)

lreArmsAll <- function(trait, arms){
	require(lre)
	data(CNLM)

	snpInfo$chrArm <- paste0(snpInfo$chrom, snpInfo$arm)
	pheno <- droplevels(pheno[!is.na(pheno[[trait]]), c("GID", "Trial", trait)])

	X <- model.matrix(~ Trial, data = pheno) 
	Z <- model.matrix(~ GID - 1, data = pheno)
	colnames(Z) <- gsub("GID", "", colnames(Z))

	M <- M[colnames(Z), ] 

	lreArms(arms = arms, snpInfo = snpInfo, y = pheno[[trait]], M = M, X = X, Z = Z)
}

# load 'doMC' library
library("doMC")

# set number of MKL threads for each parallel thread
setMKLthreads(4)

# set number of parallel threads
registerDoMC(cores = 8)

traits <- c("GY", "PH", "TW", "HD")
fitListAll <- foreach(j = traits) %:% foreach(i = 1:length(armList), .combine = c) %dopar% lreArmsAll(trait = j, arms = armList[[i]])
names(fitListAll) <- traits
save(fitListAll, "CNLMlre.RData")
